from pytorch_lightning import Trainer, LightningModule
from pytorch_lightning.callbacks import Callback
import re


class EarlyStoppingLR(Callback):

    def __init__(self, lr_threshold: float, mode="all"):
        self.lr_threshold = lr_threshold

        if mode in ("any", "all"):
            self.mode = mode
        else:
            raise ValueError(f"mode must be one of ('any', 'all')")

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._run_early_stop_checking(trainer)

    def _run_early_stop_checking(self, trainer: "pl.Trainer") -> None:
        metrics = trainer._logger_connector.callback_metrics
        if len(metrics) == 0:
            return
        all_lr = []
        for key, value in metrics.items():
            if re.match(r"opt\d+_lr\d+", key):
                all_lr.append(value)

        if len(all_lr) == 0:
            return

        if self.mode == "all":
            if all(lr <= self.lr_threshold for lr in all_lr):
                trainer.should_stop = True
        elif self.mode == "any":
            if any(lr <= self.lr_threshold for lr in all_lr):
                trainer.should_stop = True



from functools import reduce
from os.path import dirname, abspath
from typing import Callable, List, Dict, Sequence, Union

import numpy
import numpy as np
import torch
from torch import Tensor
from torch.nn import Module

from tensorneko_util.util import F, _
from tensorneko_util.util.func import generate_inf_seq, listdir, with_printed, ifelse, dict_add, as_list, \
    identity, list_to_dict, compose
from .type import T, A


def reduce_dict_by(key: str, op: Callable[[T, T], T]) -> Callable[[List[Dict[str, T]]], T]:
    """
    Apply reducer function to a list of dictionaries.

    Args:
        key (``str``): The key of the value in dicts
        op (``(T, T) -> T``): A reduce operation function used to reduce the values.

    Returns:
        ``List[Dict[str, T]] -> T``: The reducer aggregation function for the list of dictionaries.

    Examples::

        >>> x = [
        >>>     {"a": 1, "b": 2, "c": torch.Tensor([3])},
        >>>     {"a": 3, "b": 4, "c": torch.Tensor([5])},
        >>>     {"a": 2.3, "b": -1, "c": torch.Tensor([0])}
        >>> ]
        >>> reduce_dict_by("a", lambda x1, x2: x1 + x2)(x)
        6.3
        >>> reduce_dict_by("b", lambda x1, x2: x1 - x2)(x)
        -1
        >>> reduce_dict_by("c", lambda x1, x2: 10 * x1 + x2)(x)
        tensor([350.])

    """
    return F() >> (map, _[key]) >> list >> F(reduce, op)


def summarize_dict_by(key: str, op: Callable[[Union[Sequence[T], T]], T]
) -> Callable[[List[Dict[str, T]]], T]:
    """
    Apply summarizer function to a list of dictionaries.

    Args:
        key (``str``): The key of the value in dicts
        op (``(Sequence[T] | T) -> T``):
            A summarizer operation function used to summarize value from a sequence.

    Returns:
        ``List[Dict[str, T]] -> T``:
            The summarizer aggregation function for the list of dictionaries.

    Examples::

        >>> x = [
        >>>    {"a": 1, "b": torch.Tensor([2]), "c": 3, "d": np.array([1.5])},
        >>>    {"a": 3, "b": torch.Tensor([4]), "c": 5, "d": np.array([2.5])},
        >>>    {"a": 2.3, "b": torch.Tensor([-1]), "c": 0, "d": np.array([-1.0])}
        >>> ]
        >>> summarize_dict_by("a", sum)(x)
        6.3
        >>> summarize_dict_by("b", torch.mean)(x)
        tensor(1.6667)
        >>> def some_func(x):
        >>>     x = list(map(str, x))
        >>>     x = "".join(x)
        >>>     x = int(x)
        >>>     return x
        >>> summarize_dict_by("c", some_func)(x)
        350
        >>> summarize_dict_by("d", lambda x: np.sum(x, axis=0))(x)
        array([3.])

    """
    return F() >> (map, _[key]) >> list >> ifelse(
        lambda xs: type(xs[0]) is Tensor,
        func_true=torch.vstack,
        func_false=ifelse(
            lambda xs: type(xs[0]) is np.ndarray,
            func_true=numpy.vstack,
            func_false=_
        )
    ) >> op


def with_printed_shape(x: A) -> A:
    """
    An identity function but with shape printed .

    Args:
        x (:class:`~torch.Tensor` | :class:`numpy.ndarray`): Input.

    Returns:
        :class:`~torch.Tensor` | :class:`numpy.ndarray`: Identity output.

    Examples::

        >>> x = torch.tensor([1.5, 2.5, 3.5])
        >>> y = with_printed_shape(x)
        torch.Size([3])
        >>> x == y
        tensor([True, True, True])
    """
    return F(with_printed, func=lambda tensor: tensor.shape)(x)


def is_bad_num(x: Tensor) -> Tensor:
    """
    Checking the element in input tensor is bad number (NaN or Inf).

    Args:
        x (:class:`~torch.Tensor`): Input tensor.

    Returns:
        :class:`~torch.Tensor`: Checking result for bad numbers.
    """
    return torch.logical_or(torch.isnan(x), torch.isinf(x))


def count_parameters(module: Module) -> int:
    """
    Counts the number of parameters of a :class:`~torch.nn.Module`.

    Args:
        module (:class:`~torch.nn.Module`): The input module for counting parameters.

    Returns:
        ``int``: Number of parameters.
    """
    return sum(p.numel() for p in module.parameters())


def get_tensorneko_path() -> str:
    """
    Get the `tensorneko` library root path

    Returns:
        ``str``: The root path of `tensorneko`
    """
    return dirname(dirname(abspath(__file__)))


# merge package namespace
compose = compose
generate_inf_seq = generate_inf_seq
listdir = listdir
with_printed = with_printed
ifelse = ifelse
dict_add = dict_add
as_list = as_list
identity = identity
list_to_dict = list_to_dict



from typing import Optional

import torch
from torch import Tensor, zeros
from torch.nn import Parameter, LayerNorm, GELU, Identity, ModuleList

from . import MLP, ResidualBlock
from ..layer import PositionalEmbedding, Concatenate, SeqAttention
from ..neko_module import NekoModule
from ..util import ModuleFactory, compose, Shape, F


class TransformerEncoderBlock(NekoModule):
    """
    The TransformerEncoderBlock is a block in Transformer encoder which is proposed by Vaswani, et al. (2017).

    This TransformerEncoderBlock contains the multi-head attention module and the feed-forward module.

    input -> concat cls token -> add positional encoding -> AttentionModule -> FFN

    Args:
        input_shape (:class:`~tensorneko.util.type.Shape`): The shape of input sequence (N, D). N for length of
            sequence. D for embedding dimension.

        num_heads (``int``): Parallel attention heads.

        add_cls_token (``bool``, optional): The input will concat to a cls token if True. Default ``False``.

        has_pos_encoding (``bool``, optional): The input will add positional encoding if True. Default ``False``.

        linear_drop (``float``, optional): The dropout rate for linear layers. Default ``0.5``.

        attention_drop (``float``, optional): The dropout rate for attention layers. Default ``0.5``.

        build_normalization (``() -> torch.nn.Module``, optional): The normalization builder function for the block.
            Default :class:`~torch.nn.LayerNorm`.

        mlp_ratio (``float``, optional): The MLP ratio, which is the multiplication of hidden layers in FFN. e.g, if the
            embedding is 1024, and the mlp_ratio is 4.0. The FFN will be 1024 -> 4096 -> 1024. Default ``4.0``.

        build_mlp_activation (``() -> torch.nn.Module``, optional): The activation builder for FFN.
            Default :class:`~torch.nn.GELU`.

    Attributes:
        cls_token (:class:`~torch.nn.Parameter`): The trainable cls token parameters.
        pos_emb_layer (:class:`~tensorneko.layer.PositionalEmbedding`): The positional embedding layer.
        attn_module (:class:`~tensorneko.module.ResidualBlock`): The attention module with residual connection.
        feedforward_module (:class:`~tensorneko.module.ResidualBlock`): The MLP module with residual connection.

    References:
        Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
        Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

    """

    def __init__(self, input_shape: Shape, num_heads: int, add_cls_token: bool = False,
        has_pos_encoding: bool = True,
        linear_drop: float = 0.5, attention_drop: float = 0.5,
        build_normalization: Optional[ModuleFactory] = LayerNorm,
        mlp_ratio: float = 4.0, build_mlp_activation: Optional[ModuleFactory] = GELU
    ):
        super().__init__()
        # n: number of patches, d: embedding dimension
        n, d = input_shape
        # num of head
        self.num_head = num_heads
        # positional embedding
        self.add_cls_token = add_cls_token
        if self.add_cls_token:
            # prepare for class token
            self.cls_token = Parameter(zeros(1, d))
            self.token_concat = Concatenate(dim=0)
            input_shape = (n + 1, d)
        else:
            self.cls_token = None
        self.has_pos_encoding = has_pos_encoding
        if self.has_pos_encoding:
            self.pos_emb_layer = PositionalEmbedding(input_shape=input_shape, dropout_rate=linear_drop)

        # set normalization builder
        if build_normalization is LayerNorm:
            build_normalization = F(LayerNorm, input_shape)
        elif build_normalization is None:
            build_normalization = Identity

        # multi-head attention module with residual connection and normalization
        self.attn_module = ResidualBlock(
            sub_module_layers=(build_normalization, F(SeqAttention, d, self.num_head, attention_drop)),
            tail_module_layers=None
        )

        self.feedforward_module = ResidualBlock((
            build_normalization,
            F(MLP, [d, int(d * mlp_ratio), d],
              build_activation=build_mlp_activation, dropout_rate=linear_drop
              ))
        )

    def forward(self, x: Tensor) -> Tensor:
        f = F()
        if self.add_cls_token:
            f = f >> (map, lambda tokens: self.token_concat([self.cls_token, tokens])) >> list >> torch.stack
        if self.has_pos_encoding:
            f = f >> self.pos_emb_layer
        f = f >> self.attn_module >> self.feedforward_module
        return f(x)


class TransformerEncoder(NekoModule):
    """
    The TransformerEncoder repeatedly generate :class:`TransformerEncoderBlock` with specified times.

    Args:
        input_shape (:class:`~tensorneko.util.type.Shape`): The shape of input sequence (N, D). N for length of
            sequence. D for embedding dimension.

        num_heads (``int``): Parallel attention heads.

        add_cls_token (``bool``, optional): The input will concat to a cls token if True. Default ``False``.

        linear_drop (``float``, optional): The dropout rate for linear layers. Default ``0.5``.

        attention_drop (``float``, optional): The dropout rate for attention layers. Default ``0.5``.

        build_normalization (``() -> torch.nn.Module``, optional): The normalization builder function for the block.
            Default :class:`~torch.nn.LayerNorm`.

        mlp_ratio (``float``, optional): The MLP ratio, which is the multiplication of hidden layers in FFN. e.g, if the
            embedding is 1024, and the mlp_ratio is 4.0. The FFN will be 1024 -> 4096 -> 1024. Default ``4.0``.

        build_mlp_activation (``() -> torch.nn.Module``, optional): The activation builder for FFN.
            Default :class:`~torch.nn.GELU`.

        pos_encoding (``str``, optional): The option of where you want to add positional encoding. ``"all"`` for all
            blocks. ``"first"`` for first block only. ``"none"`` for no adding. Default ``"all"``.

        repeat (``int``, optional): The repeat time of TransformerEncoderBlock.

    Attributes:
        blocks (:class:`~torch.nn.ModuleList`): The blocks list in the module.

    """

    def __init__(self, input_shape: Shape, num_heads: int, add_cls_token: bool = False, linear_drop: float = 0.5,
        attention_drop: float = 0.5, build_normalization: Optional[ModuleFactory] = LayerNorm, mlp_ratio: float = 4.0,
        build_mlp_activation: Optional[ModuleFactory] = GELU, pos_encoding: str = "all",
        repeat: int = 1
    ):
        super().__init__()
        if pos_encoding == "all":
            has_pos_encoding = [True] * repeat
        elif pos_encoding == "first":
            has_pos_encoding = [True] + (repeat - 1) * [False]
        elif pos_encoding == "none":
            has_pos_encoding = [False] * repeat
        else:
            raise ValueError(f"pos_encoding must be one of 'all', 'first', 'none'. But got {pos_encoding}")

        def build_block(i):
            return TransformerEncoderBlock(
                input_shape, num_heads, add_cls_token if i == 0 else False,
                has_pos_encoding[i],
                linear_drop, attention_drop,
                build_normalization, mlp_ratio, build_mlp_activation
            )

        self.blocks = ModuleList([build_block(i) for i in range(repeat)])
        self.repeat = repeat

    def forward(self, x: Tensor) -> Tensor:
        return compose(self.blocks)(x)



from pytorch_lightning import Callback, Trainer, LightningModule


class LrLogger(Callback):
    """Log learning rate in each epoch start."""

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        for i, optimizer in enumerate(trainer.optimizers):
            for j, params in enumerate(optimizer.param_groups):
                key = f"opt{i}_lr{j}"
                value = params["lr"]
                pl_module.logger.log_metrics({key: value}, step=trainer.global_step)
                pl_module.log(key, value, logger=False, sync_dist=pl_module.distributed)



from tensorneko_util.visualization.watcher.component import Component, Variable, ProgressBar, Image, Logger, LineChart
from tensorneko_util.visualization.watcher.server import Server
from tensorneko_util.visualization.watcher.view import View

__all__ = [
    "Component",
    "Variable",
    "ProgressBar",
    "Image",
    "Logger",
    "LineChart",
    "View",
    "Server"
]



from einops.layers.torch import Rearrange as _Rearrange

from tensorneko_util.util import AbstractServer, dispatch, AverageMeter, tensorneko_util_path
from tensorneko_util.util.fp import Seq, AbstractSeq, curry, Some, Empty, F, Stream, return_option, Option, _, __
from tensorneko_util.util import ref, Timer, Singleton
from tensorneko_util.util.eventbus import Event, EventBus, EventHandler, subscribe, subscribe_async, \
    subscribe_process, subscribe_thread
from . import type
from .configuration import Configuration
from .func import reduce_dict_by, summarize_dict_by, with_printed_shape, is_bad_num, count_parameters, compose, \
    generate_inf_seq, listdir, with_printed, ifelse, dict_add, as_list, identity, list_to_dict
from .func import get_tensorneko_path
from .reproducibility import Seed
from .string_getter import get_activation, get_loss
from .type import ModuleFactory, Shape, Device

Rearrange = _Rearrange

tensorneko_path = get_tensorneko_path()
tensorneko_util_path = tensorneko_util_path

__all__ = [
    "reduce_dict_by",
    "summarize_dict_by",
    "identity",
    "generate_inf_seq",
    "compose",
    "listdir",
    "with_printed",
    "with_printed_shape",
    "is_bad_num",
    "ifelse",
    "dict_add",
    "count_parameters",
    "as_list",
    "list_to_dict",
    "ref",
    "ModuleFactory",
    "Shape",
    "Device",
    "Configuration",
    "get_activation",
    "get_loss",
    "Seed",
    "__",
    "Stream",
    "Seq",
    "return_option",
    "Option",
    "Some",
    "Empty",
    "Rearrange",
    "_",
    "F",
    "tensorneko_path",
    "tensorneko_util_path",
    "AbstractServer",
    "dispatch",
    "AverageMeter",
    "AbstractSeq",
    "curry",
    "Timer",
    "Event",
    "EventBus",
    "EventHandler",
    "subscribe",
    "subscribe_async",
    "subscribe_process",
    "subscribe_thread",
    "Singleton",
]



from abc import abstractmethod
from typing import Union, Tuple, Type

import torch
from torch import Tensor
from torch.nn import Conv2d

from .conv import _Conv, C


class _MaskedConv2d(Conv2d):
    """
    Naive masked 2D convolution layer from PixelCNN. There are 2 types (A and B). For the details, please check the
    original paper.

    For gated masked convolution, please refer to :class:`~tensorneko.module.gated_conv.GatedConv`.

    Args:
        in_channels (``int``): Number of channels in the input image

        out_channels (``int``): Number of channels produced by the convolution

        kernel_size (``int`` | ``(int, ...)``): Size of the convolving kernel

        stride (``int`` | ``(int, ...)``, optional): Stride of the convolution. Default: 1

        padding (``int`` | ``(int, ...)`` | ``str``, optional): Padding added to all four sides of
            the input. Default: 0

        dilation (``int`` | ``(int, ...)``, optional): Spacing between kernel elements. Default: 1

        groups (``int``, optional): Number of blocked connections from input
            channels to output channels. Default: 1

        bias (``bool``, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``

        padding_mode (``string``, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``

        device: The device to place the layer passed to pytorch Conv layer.

        dtype: The data type of the layer passed to pytorch Conv layer.

    References:

        van den Oord, A., Kalchbrenner, N., Espeholt, L., kavukcuoglu,  koray, Vinyals, O., & Graves, A. (2016).
        Conditional Image Generation with PixelCNN Decoders. Advances in Neural Information Processing Systems, 29.
        https://proceedings.neurips.cc/paper/2016/hash/b1301141feffabac455e1f90a7de2054-Abstract.html

    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, Tuple[int, ...]],
        stride: Union[int, Tuple[int, ...]] = 1, padding: Union[int, Tuple[int, ...], str] = 0,
        dilation: Union[int, Tuple[int, ...]] = 1, groups: int = 1, bias: bool = True, padding_mode: str = 'zeros',
        device=None, dtype=None
    ):
        super().__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode,
                         device, dtype)
        self.register_buffer("mask", self.make_mask())

    def forward(self, x: Tensor) -> Tensor:
        self.weight.data *= self.mask
        return super().forward(x)

    @abstractmethod
    def make_mask(self) -> Tensor:
        pass


class _MaskedConv2dA(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        mask[:, :, kernel_h // 2, kernel_w // 2:] = 0
        mask[:, :, kernel_h // 2 + 1:] = 0
        return mask


class _MaskedConv2dB(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        mask[:, :, kernel_h // 2, kernel_w // 2 + 1:] = 0
        mask[:, :, kernel_h // 2 + 1:] = 0
        return mask


class MaskedConv2dA(_Conv[_MaskedConv2dA]):

    @property
    def _class_name(self) -> str:
        return "tensorneko.layer.MaskedConv2dA"

    @property
    def _PtConv(self) -> Type[C]:
        return _MaskedConv2dA


class MaskedConv2dB(_Conv[_MaskedConv2dB]):

    @property
    def _class_name(self) -> str:
        return "tensorneko.layer.MaskedConv2dB"

    @property
    def _PtConv(self) -> Type[C]:
        return _MaskedConv2dB


MaskedConv2d = {
    "A": MaskedConv2dA,
    "B": MaskedConv2dB
}


class _VerticalStackConv2dA(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        print(self.weight.size())
        mask[:, :, kernel_h // 2:] = 0
        return mask


class _VerticalStackConv2dB(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        mask[:, :, kernel_h // 2 + 1:] = 0
        return mask


class _HorizontalStackConv2dA(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        mask[:, :, :, kernel_w // 2:] = 0
        return mask


class _HorizontalStackConv2dB(_MaskedConv2d):

    def make_mask(self) -> Tensor:
        mask = torch.ones(self.weight.data.shape)
        _, _, kernel_h, kernel_w = self.weight.size()
        mask[:, :, :, kernel_w // 2 + 1:] = 0
        return mask



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class Configuration(ABC):
    """
    Configuration base abstract class.

    This is designed for a functionality of using a configuration to build objects (layers, modules, etc.). Also, this
    configuration can also be saved to file and loaded from file.

    Examples:

        An example of configuration class for :class:`torch.nn.Linear`

        .. code-block:: python

            class LinearConfiguration(Configuration):

                def __init__(self, in_features, out_features, bias):
                    super().__init__()
                    self.in_features = in_features
                    self.out_features = out_features
                    self.bias = bias

                def build(self) -> torch.nn.Linear:
                    return torch.nn.Linear(self.in_features, self.out_features, self.bias)

                # using json to save and load conf
                def save(self, file_path: str) -> None:
                    attrs = {
                        "in_features": self.in_features,
                        "out_features": self.out_features,
                        "bias": self.bias
                    }
                    with open(file_path, "w", encoding="UTF-8") as file:
                        file.write(json.dumps(attrs))

                @staticmethod
                def load(file_path: str) -> LinearConfiguration:
                    with open(file_path, "r", encoding="UTF-8") as file:
                        attrs = json.loads(file.read())
                    return LinearConfiguration(attrs["in_features"], attrs["out_features"], attrs["bias"])


    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        return iter((*self.args, *self.kwargs.values()))

    @abstractmethod
    def build(self) -> Any:
        """
        A method to build an object.

        Returns:
            ``Any``: The target object of this configuration.
        """
        ...

    @abstractmethod
    def save(self, *args, **kwargs):
        """Save this configuration to file."""
        ...

    @staticmethod
    @abstractmethod
    def load(*args, **kwargs) -> Configuration:
        """
        Load this configuration from file.

        Returns:
            :class:`Configuration`: This configuration object.
        """
        ...



from abc import ABC, abstractmethod
from functools import partial
from typing import Union, Optional, Callable, Tuple, TypeVar, Type, Generic

from torch import Tensor
from torch.nn import Conv1d as PtConv1d, Conv2d as PtConv2d, Conv3d as PtConv3d, Module
from torch.nn.modules.conv import _ConvNd as PtConvNd

from ..neko_module import NekoModule
from ..util import F

C = TypeVar("C", bound=PtConvNd)


class _Conv(NekoModule, ABC, Generic[C]):
    """
    An enhanced Conv version of pytorch version with combining activation and normalization.

    Args:
        in_channels (``int``): Number of channels in the input image

        out_channels (``int``): Number of channels produced by the convolution

        kernel_size (``int`` | ``(int, ...)``): Size of the convolving kernel

        stride (``int`` | ``(int, ...)``, optional): Stride of the convolution. Default: 1

        padding (``int`` | ``(int, ...)`` | ``str``, optional): Padding added to all four sides of
            the input. Default: 0

        dilation (``int`` | ``(int, ...)``, optional): Spacing between kernel elements. Default: 1

        groups (``int``, optional): Number of blocked connections from input
            channels to output channels. Default: 1

        bias (``bool``, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``

        padding_mode (``string``, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``

        device: The device to place the layer passed to pytorch Conv layer.

        dtype: The data type of the layer passed to pytorch Conv layer.

        build_activation (``() -> torch.nn.Module``): An activation module builder to be used in the Conv layer.

        build_normalization (``() -> torch.nn.Module``): An normalization module builder to be used in the layer.

        normalization_after_activation (``bool``, optional): Set True then applying normalization after activation.
            Default ``False``.

    Attributes:
        conv: (:class:`~torch.nn._ConvNd`): The PyTorch ConvNd object.

        activation: (:class:`~torch.nn.Module`): The PyTorch activation module in the layer.

        normalization: (:class:`~torch.nn.Module`): The PyTorch normalization module in the layer.

    Examples::

        conv1d = tensorneko.layer.Conv1d(
            in_channels=256,
            out_channels=1024,
            kernel_size=(3,),
            padding=(1,),
            build_activation=torch.nn.ReLU,
            build_normalization=lambda: torch.nn.BatchNorm1d(256),
            normalization_after_activation=False
        )

        conv2d = tensorneko.layer.Conv2d(
            in_channels=256,
            out_channels=1024,
            kernel_size=(3, 3),
            padding=(1, 1),
            build_activation=torch.nn.ReLU,
            build_normalization=lambda: torch.nn.BatchNorm2d(256),
            normalization_after_activation=False
        )

        conv3d = tensorneko.layer.Conv3d(
            in_channels=256,
            out_channels=1024,
            kernel_size=(3, 3, 3),
            padding=(1, 1, 1),
            build_activation=torch.nn.ReLU,
            build_normalization=lambda: torch.nn.BatchNorm3d(256),
            normalization_after_activation=False
        )


    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, Tuple[int, ...]],
        stride: Union[int, Tuple[int, ...]] = 1, padding: Union[int, Tuple[int, ...], str] = 0,
        dilation: Union[int, Tuple[int, ...]] = 1, groups: int = 1, bias: bool = True, padding_mode: str = 'zeros',
        device=None, dtype=None,
        build_activation: Optional[Callable[[], Module]] = None,
        build_normalization: Optional[Callable[[], Module]] = None,
        normalization_after_activation: bool = False
    ):
        super().__init__()
        self.conv: C = self._PtConv(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                              stride=stride, padding=padding, dilation=dilation,
                              groups=groups, bias=bias, padding_mode=padding_mode, device=device, dtype=dtype)

        self.has_act = build_activation is not None
        if self.has_act:
            self.activation = build_activation()

        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
            self.norm_after_act = normalization_after_activation

    def forward(self, x: Tensor) -> Tensor:
        f = F() >> self.conv
        if self.has_act and self.has_norm:
            if self.norm_after_act:
                f = f >> self.activation >> self.normalization
            else:
                f = f >> self.normalization >> self.activation
        elif self.has_act and not self.has_norm:
            f = f >> self.activation
        elif not self.has_act and self.has_norm:
            f = f >> self.normalization
        return f(x)

    def _get_name(self) -> str:
        return self._class_name

    @property
    @abstractmethod
    def _class_name(self) -> str:
        pass

    @property
    @abstractmethod
    def _PtConv(self) -> Type[C]:
        pass


class Conv1d(_Conv[PtConv1d]):

    @property
    def _class_name(self) -> str:
        return "tensorneko.layer.Conv1d"

    @property
    def _PtConv(self) -> Type[C]:
        return PtConv1d


class Conv2d(_Conv[PtConv2d]):

    @property
    def _class_name(self) -> str:
        return "tensorneko.layer.Conv2d"

    @property
    def _PtConv(self) -> Type[C]:
        return PtConv2d


class Conv3d(_Conv[PtConv3d]):

    @property
    def _class_name(self) -> str:
        return "tensorneko.layer.Conv3d"

    @property
    def _PtConv(self) -> Type[C]:
        return PtConv3d


Conv = {
    "1d": Conv1d,
    "2d": Conv2d,
    "3d": Conv3d
}



from .enum import ResizeMethod, PaddingMethod, PaddingPosition
from .pad import padding_video, padding_audio
from .resize import resize_image, resize_video
from .crop import crop_with_padding
from tensorneko_util.preprocess import ffmpeg_available, frames2video

__all__ = [
    "ResizeMethod",
    "PaddingMethod",
    "PaddingPosition",
    "padding_video",
    "padding_audio",
    "resize_image",
    "resize_video",
    "crop_with_padding",
    "frames2video"
]

if ffmpeg_available:
    from tensorneko_util.preprocess import video2frames, ffmpeg_command, merge_video_audio, resample_video_fps, mp32wav
    __all__.extend(["video2frames", "ffmpeg_command", "merge_video_audio", "resample_video_fps", "mp32wav"])




from typing import Union, List, Tuple, Optional

import torch
from torch import Tensor, cat

from ..neko_module import NekoModule


class Concatenate(NekoModule):
    """
    A module version of concatenate multiple :class:`~torch.Tensor`.

    The operation used for concatenation is :func:`torch.cat`.

    Args:
        dim (``int``, optional): The dimension chosen to be concat. Default 0.
        out (:class:`~torch.Tensor`, optional): The output tensor. Default ``None``.

    Examples::

        >>> cat_dim0 = Concatenate(dim=0)
        >>> cat_dim1 = Concatenate(dim=1)
        >>> x = torch.randn(2, 3)
        >>> x
        tensor([[ 0.6580, -1.0969, -0.4614],
                [-0.1034, -0.5790,  0.1497]])
        >>> cat_dim0((x, x, x))
        tensor([[ 0.6580, -1.0969, -0.4614],
                [-0.1034, -0.5790,  0.1497],
                [ 0.6580, -1.0969, -0.4614],
                [-0.1034, -0.5790,  0.1497],
                [ 0.6580, -1.0969, -0.4614],
                [-0.1034, -0.5790,  0.1497]])
        >>> cat_dim1((x, x, x))
        tensor([[ 0.6580, -1.0969, -0.4614,  0.6580, -1.0969, -0.4614,  0.6580,
                 -1.0969, -0.4614],
                [-0.1034, -0.5790,  0.1497, -0.1034, -0.5790,  0.1497, -0.1034,
                 -0.5790,  0.1497]])

    """

    def __init__(self, dim: Optional[int] = 0, out: Optional[Tensor] = None):
        super().__init__()
        self.dim = dim
        self.out = out

    def forward(self, xs: Union[List[Tensor], Tuple[Tensor, ...]]) -> Tensor:
        return cat(xs, self.dim, out=self.out)



from typing import Sequence, Union

import torch
from torch import Tensor

from ..neko_module import NekoModule
from ..util import F


class Aggregation(NekoModule):
    """
    The torch module for aggregation.

    Args:
        mode (``str``, optional): The mode of aggregation. Default "mean".
        dim (``int`` | ``Sequence[int]``, optional): The dimension chosen to apply aggregate function. Default None.

    Examples::

        x = torch.rand(1, 16, 32, 32)
        global_avg_pooling = Aggregation("avg", dim=(1, 2, 3))
        x_pooled = max_pooling(x)

    """
    def __init__(self, mode: str = "mean", dim: Union[int, Sequence[int]] = None):
        super().__init__()
        if mode == "mean":
            self.agg_func = F(torch.mean, dim=dim)
        elif mode == "sum":
            self.agg_func = F(torch.sum, dim=dim)
        elif mode == "max":
            self.agg_func = F(torch.max, dim=dim)
        elif mode == "min":
            self.agg_func = F(torch.min, dim=dim)
        else:
            raise ValueError("Wrong mode value. It should be in [mean, sum, max, min]")

    def forward(self, x: Tensor) -> Tensor:
        return self.agg_func(x)



# import from modules
from .aggregation import Aggregation
from .attention import SeqAttention, ImageAttention
from .concatenate import Concatenate
from .conv import Conv1d, Conv2d, Conv3d, Conv
from .linear import Linear
from .log import Log
from .masked_conv2d import MaskedConv2d, MaskedConv2dA, MaskedConv2dB
from .noise import GaussianNoise
from .patching import Patching, PatchEmbedding2d, PatchEmbedding3d
from .positional_embedding import PositionalEmbedding, AbstractNonTrainablePositionalEmbedding, \
    SinCosPositionalEmbedding
from .reshape import Reshape
from .stack import Stack
from .vector_quantizer import VectorQuantizer

__all__ = [
    "Aggregation",
    "Concatenate",
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "Conv",
    "GaussianNoise",
    "ImageAttention",
    "MaskedConv2d",
    "MaskedConv2dA",
    "MaskedConv2dB",
    "Linear",
    "Log",
    "Patching",
    "PatchEmbedding2d",
    "PatchEmbedding3d",
    "PositionalEmbedding",
    "AbstractNonTrainablePositionalEmbedding",
    "SinCosPositionalEmbedding",
    "Reshape",
    "Stack",
    "SeqAttention",
    "VectorQuantizer",
]



from typing import Union, List

from torch import Tensor
from torch.nn import functional as func

from .enum import PaddingPosition, PaddingMethod, get_enum_value
from ..util import Rearrange, F


def _get_padding_pair(padding_size: int, padding_position: Union[PaddingPosition, str]) -> List[int]:
    padding_position = get_enum_value(padding_position, PaddingPosition)
    if padding_position == "tail":
        pad = [0, padding_size]
    elif padding_position == "head":
        pad = [padding_size, 0]
    elif padding_position == "average":
        padding_head = padding_size // 2
        padding_tail = padding_size - padding_head
        pad = [padding_head, padding_tail]
    else:
        raise ValueError("Wrong padding position. It should be zero or tail or average.")
    return pad


def padding_video(tensor: Tensor, target: int,
    padding_method: Union[PaddingMethod, str] = PaddingMethod.ZERO,
    padding_position: Union[PaddingPosition, str] = PaddingPosition.TAIL
) -> Tensor:
    """
    Padding a video in temporal.

    Args:
        tensor (:class:`~torch.Tensor`): Video tensor (T, C, H, W)
        target (``int``): The target temporal length of the padded video.
        padding_method (:class:`PaddingMethod`, optional): Padding method. Default zero padding.
        padding_position (:class:`PaddingPosition`, optional): Padding position. Default tail.

    Returns:
        :class:`~torch.Tensor`: The padded video.
    """
    t, c, h, w = tensor.shape
    padding_size = target - t

    padding_method = get_enum_value(padding_method, PaddingMethod)
    pad = _get_padding_pair(padding_size, padding_position)

    if padding_method == "zero":
        f = F(func.pad, pad=[0, 0, 0, 0, 0, 0] + pad)
    elif padding_method == "same":
        f = F() >> Rearrange("t c h w -> c h w t") \
            >> F(func.pad, pad=pad + [0, 0], mode="replicate") \
            >> Rearrange("c h w t -> t c h w")
    else:
        raise ValueError("Wrong padding method. It should be zero or tail or average.")

    return f(tensor)


def padding_audio(tensor: Tensor, target: int,
    padding_method: Union[PaddingMethod, str] = PaddingMethod.ZERO,
    padding_position: Union[PaddingPosition, str] = PaddingPosition.TAIL
) -> Tensor:
    """
    Padding an audio in temporal.

    Args:
        tensor (:class:`~torch.Tensor`): Audio tensor (T, C)
        target (``int``): The target temporal length of the padded audio.
        padding_method (:class:`PaddingMethod`, optional): Padding method. Default zero padding.
        padding_position (:class:`PaddingPosition`, optional): Padding position. Default tail.

    Returns:
        :class:`~torch.Tensor`: The padded audio.
    """
    t, c = tensor.shape
    padding_size = target - t

    padding_method = get_enum_value(padding_method, PaddingMethod)
    pad = _get_padding_pair(padding_size, padding_position)

    if padding_method == "zero":
        f = F(func.pad, pad=[0, 0] + pad)
    elif padding_method == "same":
        f = F() >> Rearrange("t c -> 1 c t") >> F(func.pad, pad=pad, mode="replicate") >> Rearrange("1 c t -> t c")
    else:
        raise ValueError("Wrong padding method. It should be zero or tail or average.")

    return f(tensor)



from abc import ABC
from typing import Optional, Dict

import torch
from torch import Tensor

from .gan import GAN


class WGAN(GAN, ABC):
    """
    Wasserstein Generative Adversarial Networks (WGAN) or WGAN-GP. To use this model, you need to implement
    `build_generator` and `build_discriminator` methods.

    Args:
        latent_dim (``int``): Dimension of latent space.
        g_learning_rate (``float``): Learning rate for generator.
        d_learning_rate (``float``): Learning rate for discriminator.
        gp_weight (``float``, optional): Weight of gradient penalty if value > 0. Defaults to 10.
        g_steps (``int``, optional): Number of steps for training generator per training step. Default: 1.
        d_steps (``int``, optional): Number of steps for training discriminator per training step. Default: 1.
        num_samples (``int``, optional): Number of samples for visualization in TensorBoard. Default: 8.
        grid_cols (``int``, optional): Number of columns for visualization in TensorBoard. Default: num_samples.
        name (``str``, optional): Name of the model. Default: "wgan".
        *args: Other arguments for :class:`~tensorneko.neko_model.NekoModel`.
        **kwargs: Other keyword arguments for :class:`~tensorneko.neko_model.NekoModel`.

    Attributes:
        generator (:class:`~torch.nn.Module`): Generator network.
        discriminator (:class:`~torch.nn.Module`): Discriminator network.

    References:
        Arjovsky, M., Chintala, S., & Bottou, L. (2017). Wasserstein Generative Adversarial Networks. Proceedings of the
        34th International Conference on Machine Learning, 214–223. https://proceedings.mlr.press/v70/arjovsky17a.html

        Gulrajani, I., Ahmed, F., Arjovsky, M., Dumoulin, V., & Courville, A. C. (2017). Improved Training of
        Wasserstein GANs. Advances in Neural Information Processing Systems, 30.
        https://proceedings.neurips.cc/paper/2017/hash/892c3b1c6dccd52936e27cbd0ff683d6-Abstract.html

    """

    def __init__(self, latent_dim: int, g_learning_rate: float, d_learning_rate: float, gp_weight: float = 10.0,
        g_steps: int = 1, d_steps: int = 1, num_samples: int = 8, grid_cols: Optional[int] = None,
        name: str = "wgan", *args, **kwargs
    ):
        super(WGAN, self).__init__(latent_dim, g_learning_rate, d_learning_rate, g_steps, d_steps, num_samples,
            grid_cols, name, *args, **kwargs)
        self.gp_weight = gp_weight

    @staticmethod
    def d_loss_fn(pred: Tensor, target: Tensor) -> Tensor:
        real_score = pred[target == 1]
        fake_score = pred[target == 0]
        return fake_score.mean() - real_score.mean()

    @staticmethod
    def g_loss_fn(pred: Tensor, target: Tensor) -> Tensor:
        return -pred.mean()

    def gradient_penalty_fn(self, batch_size: int, real_images: Tensor, fake_images: Tensor) -> Tensor:
        alpha = torch.randn(size=(batch_size, 1, 1, 1), device=self.device)
        diff = fake_images - real_images

        # 1. Get the discriminator output for this interpolated image.
        interpolates = torch.autograd.Variable(real_images + alpha * diff, requires_grad=True)
        disc_interpolates = self.discriminator(interpolates)

        # 2. Calculate the gradients w.r.t to this interpolated image.
        gradients = torch.autograd.grad(outputs=disc_interpolates, inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=self.device),
            create_graph=True, retain_graph=True, only_inputs=True)[0]

        # 3. Calculate the norm of the gradients.
        gradients = gradients.view(gradients.size(0), -1)
        gp = ((gradients.norm(2, dim=1) - 1) ** 2).mean()
        return gp

    def d_step(self, x: Tensor, z: Tensor) -> Dict[str, Tensor]:
        # forward discriminator
        # generate fake images
        fake_image = self.generator(z)
        fake_label = torch.zeros(x.size(0), 1, device=self.device)

        # generate real label
        real_image = x
        real_label = torch.ones(x.size(0), 1, device=self.device)

        # concat fake and real images for training
        d_batch = torch.cat((fake_image, real_image), dim=0)
        d_label = torch.cat((fake_label, real_label), dim=0)

        # calculate discriminator loss
        d_loss = self.d_loss_fn(self.discriminator(d_batch), d_label)

        # gradient penalty
        if self.training and self.gp_weight > 0:
            gp = self.gradient_penalty_fn(x.size(0), x, fake_image)
            loss = d_loss + self.gp_weight * gp
        else:
            gp = torch.tensor(0., device=self.device)
            loss = d_loss

        return {"loss": loss, "d_loss": d_loss, "gp": gp}



from typing import Tuple

import torch
from einops import rearrange
from torch import Tensor
from torch.nn import Embedding

from ..neko_module import NekoModule


class VectorQuantizer(NekoModule):
    """
    The vector quantization layer.

    Args:
        n_embeddings (``int``): The number of embeddings.
        embedding_dim (``int``): The dimension of embeddings.
        beta (``float``): The beta value for embedding loss.

    Attributes:
        embedding (:class:`~torch.nn.Embedding`): The embedding PyTorch layer.

    References:
        Oord, A., Vinyals, O., & Kavukcuoglu, K. (2017). Neural Discrete Representation Learning. Retrieved 18 May 2022,
        from https://arxiv.org/abs/1711.00937

    """

    def __init__(self, n_embeddings: int, embedding_dim: int, beta: float = 0.25):
        super().__init__()
        self.n_embedding = n_embeddings
        self.embedding_dim = embedding_dim
        self.beta = beta

        self.embedding = Embedding(self.n_embedding, self.embedding_dim)
        self.embedding.weight.data.uniform_(-1.0 / self.n_embedding, 1 / self.n_embedding)

    def forward(self, z: Tensor) -> Tuple[Tensor, Tensor]:
        z_index = self.predict_indexes(z)

        z_q = self.indexing(z_index)
        z_q = z + (z_q - z).detach()

        embedding_loss = torch.mean((z_q.detach() - z) ** 2) + self.beta * torch.mean((z_q - z.detach()) ** 2)
        return z_q, embedding_loss

    def predict_indexes(self, z: Tensor) -> Tensor:
        z = rearrange(z, "b c h w -> b h w c").contiguous()
        z_flattened = z.view(-1, self.embedding_dim)

        # calculate distance from each z to each entry
        d = torch.sum(z_flattened ** 2, dim=1, keepdim=True) + \
            torch.sum(self.embedding.weight ** 2, dim=1) - 2 * \
            torch.einsum("bd,dn->bn", z_flattened, rearrange(self.embedding.weight, "n d -> d n"))

        min_encoding_indices = torch.argmin(d, dim=1)
        return min_encoding_indices.view(z.shape[:3])

    def indexing(self, z_index: Tensor) -> Tensor:
        # z_index: (b, h, w)
        z_q = self.embedding(z_index)  # (b, h, w, c)
        return rearrange(z_q, "b h w c -> b c h w")



import re
import subprocess

import torch
from torch import Tensor
from torchmetrics.functional import structural_similarity_index_measure as ssim

from tensorneko_util.preprocess import ffmpeg_available
from tensorneko_util.util import dispatch
from .enum import Reduction
from ..io import read
from ..preprocess import padding_video, PaddingMethod


@dispatch
def ssim_image(pred: str, real: str) -> Tensor:
    """
    Calculate SSIM of an image.

    Args:
        pred (``str``): Path to the predicted image.
        real (``str``): Path to the real image.

    Returns:
        :class:`~torch.Tensor`: The ssim of the image.
    """
    pred_image = read.image(pred).unsqueeze(0)
    real_image = read.image(real).unsqueeze(0)
    return ssim(pred_image, real_image, data_range=1.0, reduction="elementwise_mean")


@dispatch
def ssim_image(pred: Tensor, real: Tensor, reduction: Reduction = Reduction.MEAN) -> Tensor:
    """
    Calculate SSIM of an image or a batch of images.

    Args:
        pred (:class:`~torch.Tensor`): Predicted images tensor. (B, C, H, W) or (C, H, W)
        real (:class:`~torch.Tensor`): Real images tensor. (B, C, H, W) or (C, H, W)
        reduction: (:class:`tensorneko.evaluation.enum.Reduction`, optional): Reduction method.
            Default: ``Reduction.MEAN``.

    Returns:
        :class:`~torch.Tensor`: The ssim of the images.
    """
    if pred.dim() == 3:
        pred = pred.unsqueeze(0)

    if real.dim() == 3:
        real = real.unsqueeze(0)

    assert pred.shape[0] == real.shape[0], "The number of images in pred and real must be equal."

    reduction_method = reduction.value

    if reduction_method == "mean":
        reduction_method = "elementwise_mean"

    return ssim(pred, real, data_range=1.0, reduction=reduction_method)


@dispatch
def ssim_video(pred: str, real: str, use_ffmpeg: bool = False) -> Tensor:
    """
    Calculate SSIM of a video.

    Args:
        pred (``str``): Path to the predicted video.
        real (``str``): Path to the real video.
        use_ffmpeg (``bool``, optional): Whether to use ffmpeg to calculate ssim.

    Returns:
        :class:`~torch.Tensor`: The ssim of the video.
    """
    if use_ffmpeg:
        if not ffmpeg_available:
            raise RuntimeError("ffmpeg is not found.")

        p = subprocess.run(["-i", pred, "-i", real, "-filter_complex", "ssim", "-f", "null", "/dev/null"],
                           capture_output=True)
        ssim_out = p.stdout.decode().split("\n")[-2]
        return Tensor(float(re.search(r" All:([\d.]+) ", ssim_out)[1]))
    else:
        pred_video = read.video(pred).video  # (T, C, H, W)
        real_video = read.video(real).video
        return ssim_video(pred_video, real_video)


@dispatch
def ssim_video(pred: Tensor, real: Tensor) -> Tensor:
    """
    Calculate SSIM of a video.

    Args:
        pred (:class:`~torch.Tensor`): Predicted video tensor. (T, C, H, W)
        real (:class:`~torch.Tensor`): Real video tensor. (T, C, H, W)

        Returns:
            ``float``: The ssim of the video.
    """
    real_video = padding_video(real, pred.shape[0], PaddingMethod.SAME)
    return ssim_image(pred, real_video, reduction=Reduction.MEAN)



from .reader import Reader
from .writer import Writer
from tensorneko_util.io import json_data, VisualLib

read = Reader()
write = Writer()

__all__ = [
    "read",
    "write",
    "json_data",
    "VisualLib"
]



from typing import Union

import torch

from ..neko_module import NekoModule
from ..util import Device


class GaussianNoise(NekoModule):
    """
    The layer for adding a Gaussian noise. Get inspired from YannDubs1 (2020).

    Args:
        sigma (``float``, optional): relative standard deviation used to generate the
            noise. Relative means that it will be multiplied by the magnitude of
            the value you are adding the noise to. This means that sigma can be
            the same regardless of the scale of the vector. Default ``0.1``.
        device (:class:`~tensorneko.util.type.Device`): The model running device. Default ``"cuda"``.

    References:
        Writing a simple Gaussian noise layer in Pytorch. (2017). Retrieved
        from https://discuss.pytorch.org/t/writing-a-simple-gaussian-noise-layer-in-pytorch/4694/3
    """

    def __init__(self, sigma=0.1, device: Union[Device, str] = "cuda"):
        super().__init__()
        self.sigma = sigma
        self.noise = torch.tensor(0.).to(device)

    def forward(self, x):
        if self.training and self.sigma != 0:
            scale = self.sigma * x.detach()
            sampled_noise = self.noise.repeat(*x.size()).normal_() * scale
            x = x + sampled_noise
        return x



from .log_graph import log_graph
from tensorneko_util.visualization import Colors, ContinuousColors, MultiPlots, tensorboard
from . import watcher
from . import matplotlib


__all__ = [
    "log_graph",
    "watcher",
    "tensorboard",
    "Colors",
    "ContinuousColors",
    "matplotlib",
    "MultiPlots"
]

try:
    from tensorneko_util.visualization.seaborn import seaborn
except ImportError:
    pass
else:
    __all__.append("seaborn")



from typing import Optional, Union, Tuple

import torch.nn.functional
from einops import rearrange
from einops.layers.torch import Rearrange
from torch import Tensor
from torch.nn import MultiheadAttention, Linear, Conv2d

from ..neko_module import NekoModule
from ..util import F


class SeqAttention(NekoModule):
    """
    The SeqAttention is the layer taking the input and calculate Q, K and V and feed them into MultiHeadAttention
    layer.

    input (Batch, Temporal, Dim) -> (Q, K, V) -> MultiHeadAttention

    The MultiHeadAttention is proposed by Vaswani, et al. (2017).

    Args:
        embed_dim (``int``): The embedding dim of input sequence.

        num_heads (``bool``): Parallel attention heads.

        dropout (``float``, optional): A Dropout layer on attn_output_weights. Default: 0.0.

        bias (``bool``, optional): Add bias as module parameter. Default: True.

        add_bias_kv (``bool``, optional): Add bias to the key and value sequences at dim=0.

        add_zero_attn (``bool``, optional): Add a new batch of zeros to the key and value sequences at dim=1.

        kdim (``int``, optional): Total number of features in key. Default: None.

        vdim (``int``, optional): Total number of features in value. Default: None.

        return_attention_weights (``bool``, optional): Return both value output and attention weights if True else
            return attention value output only.

    Attributes:
        q_linear (:class:`~torch.nn.Linear`): The PyTorch Linear layer for calculating Q.
        k_linear (:class:`~torch.nn.Linear`): The PyTorch Linear layer for calculating K.
        v_linear (:class:`~torch.nn.Linear`): The PyTorch Linear layer for calculating V.
        attention (:class:`~torch.nn.MultiheadAttention`): The PyTorch MultiheadAttention layer of this module.

    References:
        Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
        Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

    """

    def __init__(self, embed_dim: int, num_heads: int, dropout: float = 0., bias: bool = True,
        add_bias_kv: bool = False, add_zero_attn: bool = False, kdim: Optional[int] = None, vdim: Optional[int] = None,
        return_attention_weights: bool = False
    ):
        super().__init__()
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.q_linear = Linear(embed_dim, embed_dim)
        self.k_linear = Linear(embed_dim, self.kdim)
        self.v_linear = Linear(embed_dim, self.vdim)
        self.attention = MultiheadAttention(embed_dim, num_heads, dropout, bias, add_bias_kv, add_zero_attn, kdim, vdim,
            batch_first=True)
        self.return_attention_weights = return_attention_weights

    def forward(self, x: Tensor) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        f = F() >> (map, lambda linear: linear(x)) >> (lambda xs: self.attention(*xs))
        x, weight = f([self.q_linear, self.k_linear, self.v_linear])
        return (x, weight) if self.return_attention_weights else x


class ImageAttention(NekoModule):
    """
    The ImageAttentionModule is the layer taking the input and calculate Q, K and V and feed them into
    MultiHeadAttention layers.

    input (Batch, Channel, Height, Width) -> (Q, K, V) -> MultiHeadAttention

    The main difference of this class to :class:`~tensorneko.layer.attention.SeqAttention` is the input is a batch of
    feature maps.

    The MultiHeadAttention is proposed by Vaswani, et al. (2017).

    Args:
        embed_dim (``int``): The embedding dim of input sequence.

        num_heads (``bool``): Parallel attention heads.

        dropout (``float``, optional): A Dropout layer on attn_output_weights. Default: 0.0.

        bias (``bool``, optional): Add bias as module parameter. Default: True.

        add_bias_kv (``bool``, optional): Add bias to the key and value sequences at dim=0.

        add_zero_attn (``bool``, optional): Add a new batch of zeros to the key and value sequences at dim=1.

        kdim (``int``, optional): Total number of features in key. Default: None.

        vdim (``int``, optional): Total number of features in value. Default: None.

        return_attention_weights (``bool``, optional): Return both value output and attention weights if True else
            return attention value output only.

    Attributes:
        q_linear (:class:`~torch.nn.Conv2d`): The PyTorch Conv2d layer for calculating Q.
        k_linear (:class:`~torch.nn.Conv2d`): The PyTorch Conv2d layer for calculating K.
        v_linear (:class:`~torch.nn.Conv2d`): The PyTorch Conv2d layer for calculating V.
        attention (:class:`~torch.nn.MultiheadAttention`): The PyTorch MultiheadAttention layer of this module.

    References:
        Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
        Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

    """

    def __init__(self, embed_dim: int, num_heads: int, dropout: float = 0., bias: bool = True,
        add_bias_kv: bool = False, add_zero_attn: bool = False, kdim: Optional[int] = None, vdim: Optional[int] = None,
        return_attention_weights: bool = False
    ):
        super().__init__()
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.q_linear = Conv2d(embed_dim, embed_dim, 1)
        self.k_linear = Conv2d(embed_dim, self.kdim, 1)
        self.v_linear = Conv2d(embed_dim, self.vdim, 1)
        self.attention = MultiheadAttention(embed_dim, num_heads, dropout, bias, add_bias_kv, add_zero_attn, kdim, vdim,
            batch_first=True)
        self.return_attention_weights = return_attention_weights

    def forward(self, x: Tensor) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        _, _, h, w = x.size()
        f = F() >> (map, lambda linear: linear(x)) >> (map, Rearrange("b c h w -> b (h w) c")) \
            >> (lambda xs: self.attention(*xs))
        x, weight = f([self.q_linear, self.k_linear, self.v_linear])
        x = rearrange(x, "b (h w) c -> b c h w", h=h, w=w)
        return (x, weight) if self.return_attention_weights else x



from .display_metrics_callback import DisplayMetricsCallback
from .lr_logger import LrLogger
from .epoch_num_logger import EpochNumLogger
from .epoch_time_logger import EpochTimeLogger
from .nil_callback import NilCallback
from .earlystop_lr import EarlyStoppingLR

__all__ = [
    "LrLogger",
    "EpochNumLogger",
    "EpochTimeLogger",
    "NilCallback",
    "DisplayMetricsCallback",
    "EarlyStoppingLR",
]



from __future__ import annotations

from typing import Type

from tensorneko_util.io.writer import Writer as BaseWriter

try:
    from .mesh import MeshWriter
    pytorch3d_available = True
except ImportError:
    MeshWriter = object
    pytorch3d_available = False


class Writer(BaseWriter):

    def __init__(self):
        super().__init__()
        self._mesh = None

    @property
    def mesh(self) -> Type[MeshWriter]:
        if pytorch3d_available:
            if self._mesh is None:
                from .mesh import MeshWriter
                self._mesh = MeshWriter()
            return self._mesh
        else:
            raise ImportError("To use the mesh writer, please install pytorch3d.")



from abc import abstractmethod
from typing import Optional, Dict, Union, Sequence, Any

import torch
from pytorch_lightning import LightningModule
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.utilities.types import EPOCH_OUTPUT, STEP_OUTPUT
from torch import Tensor

from .neko_module import NekoModule
from .neko_trainer import NekoTrainer
from .util import summarize_dict_by, Shape


class NekoModel(LightningModule, NekoModule):
    """
    An abstract class for models. In this module, the loss and other metrics will be automatically logged in
    TensorBoard.

    Args:
        name (``str``): Model name

        input_shape (:class:`~tensorneko.util.Shape`, optional):
            An optional argument can allow it to plot a graph for TensorBoard

        *args: Other arguments for :class:`~pytorch_lightning.core.lightning.LightningModule`

        **kwargs: Other arguments for :class:`~pytorch_lightning.core.lightning.LightningModule`
    """
    trainer: NekoTrainer

    def __init__(self, name: str,
        input_shape: Optional[Shape] = None, distributed=False, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.name = name
        self.distributed = distributed
        self.can_plot_graph = input_shape is not None
        if self.can_plot_graph:
            self.input_shape = input_shape
            self.example_input_array = torch.rand([1, *input_shape])
        self.history = []

    @abstractmethod
    def forward(self, *args, **kwargs):
        ...

    @abstractmethod
    def training_step(self,
        batch: Optional[Union[Tensor, Sequence[Tensor]]] = None,
        batch_idx: Optional[int] = None,
        optimizer_idx: Optional[int] = None,
        hiddens: Optional[Tensor] = None
    ) -> Dict[str, Tensor]:
        """
        The method inherit from :meth:`~pytorch_lightning.core.lightning.LightningModule.training_step`.

        Here you compute and return the training loss and some additional metrics for e.g.
        the progress bar or logger.

        Args:
            batch (:class:`~torch.Tensor` | (:class:`~torch.Tensor`, ...) | [:class:`~torch.Tensor`, ...]):
                The output of your :class:`~torch.utils.data.DataLoader`. A tensor, tuple or list.

            batch_idx (``int``): Integer displaying index of this batch

            optimizer_idx (``int``, optional): When using multiple optimizers, this argument will also be present.

            hiddens(:class:`~torch.Tensor`, optional): Passed in if
                :paramref:`~pytorch_lightning.core.lightning.LightningModule.truncated_bptt_steps` > 0.

        Returns:
            ``Dict`` [``str``, :class:`~torch.Tensor`]:
                A dictionary. Can include any keys, but must include the key ``'loss'``

        Notes:
            The return value can be other types, but you need to handle them by override the
                method :meth:`~pytorch_lightning.TrainingModule.training_step_end`

        Examples::

            def training_step(self, batch, batch_idx):
                x, y = batch
                out = self(x)
                loss = self.loss(out, x)
                return {"loss": loss}
        """
        ...

    @abstractmethod
    def validation_step(self,
        batch: Optional[Union[Tensor, Sequence[Tensor]]] = None,
        batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:
        """
        The method inherit from :meth:`~pytorch_lightning.core.lightning.LightningModule.validation_step`.

        Operates on a single batch of data from the validation set.
        In this step you'd might generate examples or calculate anything of interest like accuracy.

        Args:
            batch (:class:`~torch.Tensor` | (:class:`~torch.Tensor`, ...) | [:class:`~torch.Tensor`, ...]):
                The output of your :class:`~torch.utils.data.DataLoader`. A tensor, tuple or list.

            batch_idx (``int``): The index of this batch

            dataloader_idx (``int``, optional): The index of the dataloader that produced this batch
                (only if multiple val dataloaders used)

        Returns:
            ``Dict`` [``str``, :class:`~torch.Tensor`]:
                A dictionary. Can include any keys, but must include the key ``'loss'``

        Examples::

            def validation_step(self, batch, batch_idx):
                x, y = batch
                out = self(x)
                loss = self.loss(out, x)
                acc = self.acc(out, x)
                return {"loss": loss, "acc": acc}
        """
        ...

    def predict_step(self, batch: Tensor, batch_idx: int, dataloader_idx: Optional[int] = None) -> Any:
        """
        Inherit from :meth:`~pytorch_lightning.core.lightning.LightningModule.predict_step`.

        Step function called during :meth:`~pytorch_lightning.trainer.trainer.Trainer.predict`.
        By default, it calls :meth:`~pytorch_lightning.core.lightning.LightningModule.forward`.
        Override to add any processing logic.

        Args:
            batch (:class:`~torch.Tensor`): Current batch
            batch_idx (``int``): Index of current batch
            dataloader_idx (``int``, optional): Index of the current dataloader

        Return:
            ``Any``: The predicted output
        """
        return self.forward(batch)

    def test_step(self,
        batch: Optional[Union[Tensor, Sequence[Tensor]]] = None,
        batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:
        return self.validation_step(batch, batch_idx, dataloader_idx)

    @abstractmethod
    def configure_optimizers(self):
        """Inherit from :meth:`~pytorch_lightning.core.lightning.LightningModule.configure_optimizers`."""
        ...

    @property
    def logger(self) -> Optional[LightningLoggerBase]:
        """
        Returns:
            :class:`~pytorch_lightning.loggers.LightningLoggerBase` | ``None``:
                The training or validation logger for the module.
        """
        if not self.trainer:
            return None
        else:
            return self.trainer.logger

    def training_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        """For each training epoch end, log the metrics"""
        if self.trainer.log_on_epoch and self.logger is not None:
            self.log_on_training_epoch_end(outputs)

    def validation_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        """For each validation epoch end, log the metrics"""
        if self.logger is not None:
            self.log_on_validation_epoch_end(outputs)

    def training_step_end(self, output: STEP_OUTPUT) -> STEP_OUTPUT:
        """For each training step end, log the metrics"""
        if self.trainer.log_on_step \
            and self.logger is not None \
            and self.trainer.global_step % self.trainer.log_every_n_steps == 0:
            self.log_on_training_step_end(output)
        return output

    def log_on_training_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        """Log the training epoch outputs"""
        history_item = {}
        for key in outputs[0].keys():
            getter = summarize_dict_by(key, torch.mean)
            value = getter(outputs)
            history_item[key] = value
            self.logger.log_metrics({key: value}, step=self.trainer.global_step)
            self.log(key, value, on_epoch=True, on_step=False, logger=False, sync_dist=self.distributed)
        self.history.append(history_item)

    def log_on_validation_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        """Log the validation epoch outputs"""
        if len(self.history) == 0 or "loss" not in self.history[-1].keys():
            return

        for key in outputs[0].keys():
            getter = summarize_dict_by(key, torch.mean)
            value = getter(outputs)
            self.history[-1]["val_" + key] = value
            self.logger.log_metrics({key: value}, step=self.trainer.global_step)
            self.log(key, value, on_epoch=True, on_step=False, logger=False, sync_dist=self.distributed)
            self.log(f"val_{key}", value, on_epoch=True, on_step=False, logger=False, prog_bar=True,
                     sync_dist=self.distributed)

    def log_on_training_step_end(self, output: STEP_OUTPUT) -> None:
        """Log the training step outputs"""
        history_item = {}
        for key, value in output.items():
            history_item[key] = value
            self.logger.log_metrics({key: value}, step=self.trainer.global_step)
            self.log(key, value, on_epoch=False, on_step=True, logger=False, sync_dist=self.distributed)
        self.history.append(history_item)

    def test_step_end(self, output: STEP_OUTPUT) -> Optional[STEP_OUTPUT]:
        """For each test step end, log the metrics"""
        self.log_dict(output, sync_dist=self.distributed)
        return output

    def log_image(self, name: str, image: torch.Tensor) -> None:
        """Log an image to the logger"""
        if self.logger is None:
            return

        self.logger.experiment.add_image(name, torch.clip(image, 0, 1), self.trainer.global_step)



from abc import abstractmethod, ABC
from typing import Optional, Union, Sequence, Dict

import torch
import torchvision
from torch import Tensor
from torch.nn import Module
from torch.nn.functional import binary_cross_entropy_with_logits
from torch.optim import Adam

from ..neko_model import NekoModel


class GAN(NekoModel, ABC):
    """
    Vanilla Generative Adversarial Networks (GAN). To use this model, you need to implement `build_generator` and
    `build_discriminator` methods.

    Args:
        latent_dim (``int``): Dimension of latent space.
        g_learning_rate (``float``): Learning rate for generator.
        d_learning_rate (``float``): Learning rate for discriminator.
        g_steps (``int``, optional): Number of steps for training generator per training step. Default: 1.
        d_steps (``int``, optional): Number of steps for training discriminator per training step. Default: 1.
        num_samples (``int``, optional): Number of samples for visualization in TensorBoard. Default: 8.
        grid_cols (``int``, optional): Number of columns for visualization in TensorBoard. Default: num_samples.
        name (``str``, optional): Name of the model. Default: "gan".
        *args: Other arguments for :class:`~tensorneko.neko_model.NekoModel`.
        **kwargs: Other keyword arguments for :class:`~tensorneko.neko_model.NekoModel`.

    Attributes:
        generator (:class:`~torch.nn.Module`): Generator network.
        discriminator (:class:`~torch.nn.Module`): Discriminator network.

    References:
        Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., Courville, A., & Bengio, Y.
        (2014). Generative Adversarial Nets. Advances in Neural Information Processing Systems, 27.
        https://proceedings.neurips.cc/paper/2014/hash/5ca3e9b122f61f8f06494c97b1afccf3-Abstract.html

    """

    def __init__(self, latent_dim: int, g_learning_rate: float, d_learning_rate: float,
        g_steps: int = 1, d_steps: int = 1, num_samples: int = 8, grid_cols: Optional[int] = None,
        name: str = "gan", *args, **kwargs
    ):
        super().__init__(name, *args, **kwargs)

        self.latent_dim = latent_dim
        self.g_learning_rate = g_learning_rate
        self.d_learning_rate = d_learning_rate
        self.g_steps = g_steps
        self.d_steps = d_steps
        self.sample_z = None
        self.num_samples = num_samples
        self.grid_cols = grid_cols or num_samples

        self.generator = self.build_generator()
        self.discriminator = self.build_discriminator()
        self.automatic_optimization = False

    def forward(self, z: Tensor) -> Tensor:
        return self.generator(z)

    @abstractmethod
    def build_generator(self) -> Module:
        pass

    @abstractmethod
    def build_discriminator(self) -> Module:
        pass

    @staticmethod
    def g_loss_fn(pred, target) -> Tensor:
        return binary_cross_entropy_with_logits(pred, target)

    @staticmethod
    def d_loss_fn(pred, target) -> Tensor:
        return binary_cross_entropy_with_logits(pred, target)

    def training_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        hiddens: Optional[Tensor] = None) -> Dict[str, Tensor]:
        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]

        d_optimizer, g_optimizer = self.optimizers()
        schedulers = self.lr_schedulers()
        if schedulers is not None:
            d_scheduler, g_scheduler = schedulers
        else:
            d_scheduler, g_scheduler = None, None

        # sample noise
        z = torch.randn(x.size(0), self.latent_dim, device=self.device)

        # train discriminator
        d_loss = None
        d_result = None
        for _ in range(self.d_steps):
            d_result = self.d_step(x, z)
            d_loss = d_result["loss"]
            d_optimizer.zero_grad()
            self.manual_backward(d_loss)
            d_optimizer.step()
        if d_scheduler is not None:
            d_scheduler.step()

        # train generator
        g_loss = None
        g_result = None
        for _ in range(self.g_steps):
            g_result = self.g_step(x, z)
            g_loss = g_result["loss"]
            g_optimizer.zero_grad()
            self.manual_backward(g_loss)
            g_optimizer.step()
        if g_scheduler is not None:
            g_scheduler.step()

        return {
            "loss": d_loss + g_loss,
            **{k: v for k, v in d_result.items() if k != "loss"},
            **{k: v for k, v in g_result.items() if k != "loss"},
        }

    def d_step(self, x: Tensor, z: Tensor) -> Dict[str, Tensor]:
        # forward discriminator
        # generate fake images
        fake_image = self.generator(z)
        fake_label = torch.zeros(x.size(0), 1, device=self.device)

        # generate real label
        real_image = x
        real_label = torch.ones(x.size(0), 1, device=self.device)

        # concat fake and real images for training
        d_batch = torch.cat((fake_image, real_image), dim=0)
        d_label = torch.cat((fake_label, real_label), dim=0)

        # calculate discriminator loss
        d_loss = self.d_loss_fn(self.discriminator(d_batch), d_label)
        return {"loss": d_loss, "d_loss": d_loss}

    def g_step(self, x: Tensor, z: Tensor) -> Dict[str, Tensor]:
        # forward generator
        # generate fake images and labels
        fake_image = self.generator(z)
        target_label = torch.ones(x.size(0), 1, device=self.device)  # target label is real.

        # calculate generator loss
        g_loss = self.g_loss_fn(self.discriminator(fake_image), target_label)
        return {"loss": g_loss, "g_loss": g_loss}

    def validation_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:

        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]

        # sample noise
        z = torch.randn(x.size(0), self.latent_dim, device=self.device)

        # test discriminator
        d_result = self.d_step(x, z)
        d_loss = d_result["loss"]

        # test generator
        g_result = self.g_step(x, z)
        g_loss = g_result["loss"]

        return {
            "loss": d_loss + g_loss,
            **{k: v for k, v in d_result.items() if k != "loss"},
            **{k: v for k, v in g_result.items() if k != "loss"},
        }

    def on_validation_epoch_end(self) -> None:
        super().on_epoch_end()
        if self.sample_z is None:
            self.sample_z = torch.randn(self.num_samples, self.latent_dim, device=self.device)
        sample_generated = self.generator(self.sample_z)
        grid = torchvision.utils.make_grid(sample_generated, nrow=self.grid_cols)
        self.log_image("sample_images", grid)

    def configure_optimizers(self):
        d_optimizer = Adam(self.discriminator.parameters(), lr=self.d_learning_rate)
        g_optimizer = Adam(self.generator.parameters(), lr=self.g_learning_rate)
        return [d_optimizer, g_optimizer]



from __future__ import annotations

from torch import Tensor
from torch.nn import ModuleList, Sequential, Module

from ..layer import Concatenate
from ..neko_module import NekoModule
from ..util import F


class InceptionModule(NekoModule):
    """
    InceptionModule is the module has multiple route for input feature and then concat the outputs of all routes. This
    is introduced firstly by Szegedy, et al. (2015).

    Attributes:
        sub_seqs (:class:`torch.nn.ModuleList`): The module list of sub sequential modules.

    References:
        Szegedy, C., Liu, W., Jia, Y., Sermanet, P., Reed, S., Anguelov, D., ... & Rabinovich, A. (2015). Going deeper
        with convolutions. In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 1-9).

    """

    def __init__(self):
        super().__init__()
        self.sub_seqs = ModuleList()
        self.channel_concat = Concatenate(dim=1)

    def add_sub_sequence(self, *layers: Module) -> InceptionModule:
        """
        Add sub sequential module for this InceptionModule.

        Args:
            *layers (:class:`~torch.nn.Module`): The var length parameter layers are the modules used in one route.

        Returns:
            :class:`InceptionModule`: The self InceptionModule object.

        Examples:

            Add a route with 1x1Conv -> 3x3Conv layers for this module.

            .. code-block:: python

                inception_module.add_sub_sequence(
                    tensorneko.layer.Conv2d(
                        in_channels=64,
                        out_channels=64,
                        kernel_size=(1, 1),
                        padding=(0, 0),
                        build_activation=torch.nn.ReLU,
                        build_normalization=lambda: torch.nn.BatchNorm2d(64),
                    ),
                    tensorneko.layer.Conv2d(
                        in_channels=64,
                        out_channels=64,
                        kernel_size=(3, 3),
                        padding=(1, 1),
                        build_activation=torch.nn.ReLU,
                        build_normalization=lambda: torch.nn.BatchNorm2d(64),
                    ),
                )

        """
        self.sub_seqs.append(Sequential(*layers))
        return self

    def forward(self, x: Tensor) -> Tensor:
        f = F() >> (map, lambda seq: seq(x)) >> list >> self.channel_concat
        return f(self.sub_seqs)




from time import time

from pytorch_lightning import Callback, Trainer, LightningModule


class EpochTimeLogger(Callback):
    """Log spent time for each training epoch"""
    start_time: float

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.start_time = time()

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        elapsed_time = time() - self.start_time
        pl_module.logger.log_metrics({"epoch_time": elapsed_time}, step=trainer.global_step)
        pl_module.log("epoch_time", elapsed_time, logger=False, sync_dist=pl_module.distributed)




from typing import Union

from einops import rearrange
from matplotlib import pyplot as plt
from numpy import ndarray
from torch import Tensor

from tensorneko_util.visualization.matplotlib import plot_image


def imshow(tensor: Union[Tensor, ndarray], *args, **kwargs):
    """
    Use matplotlib to show image of tensor with shape (C, H, W).

    Except the first parameter, other parameters are from :func:`matplotlib.pyplot.imshow`.

    Args:
        tensor (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The image with (C, H, W) to show.

        *args: Other parameters for :func:`matplotlib.pyplot.plot`.
        **kwargs: Other parameters for :func:`matplotlib.pyplot.imshow`.

    """
    if type(tensor) is Tensor:
        tensor = tensor.detach().cpu().numpy()
    tensor = rearrange(tensor, "c h w -> h w c")
    plt.imshow(tensor, *args, **kwargs)
    plt.show()


def plot(tensor: Union[Tensor, ndarray], *args, **kwargs):
    """
    Use matplotlib to plot tensor with shape (N, 2). First column is X and second column is Y.

    Except the first parameter, other parameters are from :func:`matplotlib.pyplot.plot`.

    Args:
        tensor (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The tensor (N, 2) to plot.

        *args: Other parameters for :func:`matplotlib.pyplot.plot`.
        **kwargs: Other parameters for :func:`matplotlib.pyplot.plot`.

    """
    if type(tensor) is Tensor:
        tensor = tensor.detach().cpu().numpy()
    plt.plot(tensor[:, 0], tensor[:, 1], *args, **kwargs)
    plt.show()


plot_image = plot_image



from typing import Iterable

import torch
from torch import randn
from torch.nn import Module
from torch.utils.tensorboard import SummaryWriter

from ..util import Shape


def log_graph(module: Module, input_shape_list: Iterable[Shape], log_dir: str = None) -> None:
    """
    Plot a graph of model for TensorBoard.

    Args:
        module (:class:`~torch.nn.Module`): The module you want to plot the structure.
        input_shape_list (``Iterable`` [:class:`~tensorneko.util.type.Shape`]): The shapes of input variables.
        log_dir (``str``, optional): The log directory for save. Default "None".

    Examples:

        Make a structure graph of a module of :class:`tensorneko.layer.Linear`.

        .. code-block:: python

            linear = tensorneko.layer.Linear(
                in_features=128,
                out_features=256,
                build_activation=torch.nn.LeakyReLU,
                build_normalization=lambda: torch.nn.BatchNorm1d(256),
                dropout_rate=0.5
            )
            log_graph(linear, [(128,)], log_dir="graphs")

        Then, run the TensorBoard server in terminal. And check the graph in TensorBoard.

        .. code-block:: bash

            tensorboard --logdir="graphs"

    """

    writer = SummaryWriter(log_dir=log_dir)
    xs = list(map(lambda shape: randn(torch.Size((1, *shape))), input_shape_list))
    if len(xs) == 1:
        writer.add_graph(module, xs[0])
    else:
        writer.add_graph(module, xs)
    writer.close()



from typing import Tuple

import numpy as np
import torch
from PIL import Image
from einops import rearrange
from numpy import ndarray
from torch import Tensor, uint8, float32
from torch.nn import functional as func

from .enum import ResizeMethod, get_enum_value
from ..util import F, _


def resize_image(tensor: Tensor, size: Tuple[int, int], resize_method: ResizeMethod = ResizeMethod.BICUBIC
) -> Tensor:
    """
    Resizing an image to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Image tensor (C, H, W) with value range [0, 1].
        size ((``int``, ``int``)): Target size (H, W).
        resize_method (:class:`ResizeMethod`, optional): Resize method. Default bicubic.

    Returns:
        :class:`~torch.Tensor`: The resized image.
    """
    h, w = size

    def resizer(x: ndarray) -> ndarray:
        x = Image.fromarray(x)
        x = x.resize((w, h), resample=resize_method.value)
        x = np.asarray(x).astype(np.uint8)
        return x

    f = F(rearrange, pattern="c h w -> h w c") \
        >> _ * 255 \
        >> (lambda x: x.to(uint8).numpy()) \
        >> resizer \
        >> torch.from_numpy \
        >> F(rearrange, pattern="h w c -> c h w") \
        >> _ / 255 \
        >> (lambda x: x.to(float32))
    return f(tensor)


def resize_video(tensor: Tensor, size: Tuple[int, int], resize_method: ResizeMethod = ResizeMethod.BICUBIC,
    fast=False
) -> Tensor:
    """
    Resizing a video to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Video tensor (T, C, H, W)
        size ((``int``, ``int``)): Target size (H, W)
        resize_method (:class:`ResizeMethod`, optional): Resize method. Default bicubic.
        fast (bool, optional): If True, use fast mode (pytorch F.interpolate). Default False.

    Returns:
        :class:`~torch.Tensor`: The resized video.
    """
    if not fast:
        image_resizer = F(resize_image, size=size, resize_method=resize_method) \
                        >> F(rearrange, pattern="c h w -> 1 c h w")
        f = F(map, image_resizer) \
            >> list \
            >> torch.vstack
        return f(tensor)
    else:
        return func.interpolate(tensor, size, mode=get_enum_value(resize_method, ResizeMethod), align_corners=False)
